import numpy

from .._io.output import read as read_output
from .._io.output import write as write_output
from .._mesh import read as read_mesh

__all__ = [
    "extract",
]


format_to_ext = {
    "csv": ".csv",
    "tecplot": ".tec",
    "petrasim": ".csv",
}


def extract(argv=None):
    import os

    parser = _get_parser()
    args = parser.parse_args(argv)

    # Check output file format
    if args.connection and args.file_format != "csv":
        raise ValueError("Connection data can only be exported to CSV.")

    # Check that TOUGH output and MESH file exist
    if not os.path.isfile(args.infile):
        raise ValueError("TOUGH output file '{}' not found.".format(args.infile))
    if not os.path.isfile(args.mesh):
        raise ValueError("MESH file '{}' not found.".format(args.mesh))

    # Read MESH and extract X, Y and Z
    parameters = read_mesh(
        args.mesh, file_format="tough", label_length=args.label_length
    )
    if "elements" not in parameters.keys():
        raise ValueError("Invalid MESH file '{}'.".format(args.mesh))

    # Read TOUGH output file
    output = read_output(
        args.infile, connection=args.connection, label_length=args.label_length
    )
    if output[-1].format != "tough":
        raise ValueError("Invalid TOUGH output file '{}'.".format(args.infile))

    try:
        if not args.connection:
            points = numpy.vstack(
                [parameters["elements"][label]["center"] for label in output[-1].labels]
            )
        else:
            points = numpy.array(
                [
                    numpy.mean(
                        [parameters["elements"][l]["center"] for l in label], axis=0
                    )
                    for label in output[-1].labels
                ]
            )
        points = {k: v for k, v in zip(["X", "Y", "Z"], points.T)}
        for out in output:
            out.data.update(points)

    except KeyError:
        raise ValueError(
            "Elements in '{}' and '{}' are not consistent.".format(
                args.infile, args.mesh
            )
        )

    # Write TOUGH3 element output file
    ext = format_to_ext[args.file_format]
    filename = (
        args.output_file
        if args.output_file is not None
        else "OUTPUT_ELEME{}".format(ext)
        if not args.connection
        else "OUTPUT_CONNE{}".format(ext)
    )
    if not args.split or len(output) == 1:
        write_output(filename, output, file_format=args.file_format)
    else:
        head, ext = os.path.splitext(filename)
        for i, out in enumerate(output):
            write_output("{}_{}{}".format(head, i + 1, ext), out, file_format="csv")


def _get_parser():
    import argparse

    # Initialize parser
    parser = argparse.ArgumentParser(
        description=(
            "Extract results from TOUGH main output file and reformat as a TOUGH3 CSV output file."
        ),
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # Input file
    parser.add_argument(
        "infile",
        type=str,
        help="TOUGH output file",
    )

    # Mesh file
    parser.add_argument(
        "mesh",
        type=str,
        help="TOUGH MESH file (can be INFILE)",
    )

    # Output file
    parser.add_argument(
        "--output-file",
        "-o",
        type=str,
        default=None,
        help="TOUGH3 element output file",
    )

    # File format
    parser.add_argument(
        "--file-format",
        "-f",
        type=str,
        choices=("csv", "tecplot", "petrasim"),
        default="csv",
        help="exported file format",
    )

    # Label length
    parser.add_argument(
        "--label-length",
        "-l",
        type=int,
        choices=(5, 6, 7, 8, 9),
        default=None,
        help="number of characters in cell labels",
    )

    # Split or not
    parser.add_argument(
        "--split",
        "-s",
        default=False,
        action="store_true",
        help="write one file per time step",
    )

    # Read connection data
    parser.add_argument(
        "--connection",
        "-c",
        default=False,
        action="store_true",
        help="extract data related to connections",
    )

    return parser



from copy import deepcopy

import meshio
import numpy

from .._common import filetype_from_filename, register_format
from ._mesh import from_meshio

__all__ = [
    "register",
    "read",
    "write",
    "read_time_series",
    "write_time_series",
]


_extension_to_filetype = {}
_reader_map = {}
_writer_map = {}
_materials = ["material", "gmsh:physical", "medit:ref"]


def register(file_format, extensions, reader, writer=None, material=None):
    """
    Register a new mesh format.

    Parameters
    ----------
    file_format : str
        File format to register.
    extensions : array_like
        List of extensions to associate to the new format.
    reader : callable
        Read fumction.
    writer : callable or None, optional, default None
        Write function.
    material : str or None, optional, default None
        Cell data key associated to material.

    """
    register_format(
        fmt=file_format,
        ext_to_fmt=_extension_to_filetype,
        reader_map=_reader_map,
        writer_map=_writer_map,
        extensions=extensions,
        reader=reader,
        writer=writer,
    )
    if material:
        _materials.append(material)


def get_material_key(cell_data):
    """Get key of material data in cell_data."""
    for k in cell_data.keys():
        if k in _materials:
            return k
    return None


def read(filename, file_format=None, **kwargs):
    """
    Read unstructured mesh from file.

    Parameters
    ----------
    filename : str
        Input file name.
    file_format : str or None, optional, default None
        Input file format.

    Other Parameters
    ----------------
    label_length : int or None, optional, default None
        Only if ``file_format = "tough"``. Number of characters in cell labels.

    Returns
    -------
    toughio.Mesh
        Imported mesh.

    """
    # Check file format
    if not isinstance(filename, str):
        raise TypeError()
    fmt = (
        file_format
        if file_format
        else filetype_from_filename(filename, _extension_to_filetype)
    )

    # Call custom readers
    if fmt in _reader_map.keys():
        mesh = _reader_map[fmt](filename, **kwargs)
        if fmt not in {"tough", "pickle"}:
            mesh.cell_data = {
                k: numpy.concatenate(v) for k, v in mesh.cell_data.items()
            }
            key = get_material_key(mesh.cell_data)
            if key:
                mesh.cell_data["material"] = mesh.cell_data.pop(key)
    else:
        mesh = meshio.read(filename, file_format)
        mesh = from_meshio(mesh)
    return mesh


def write(filename, mesh, file_format=None, **kwargs):
    """
    Write unstructured mesh to file.

    Parameters
    ----------
    filename : str
        Output file name.
    mesh : toughio.Mesh
        Mesh to export.
    file_format : str or None, optional, default None
        Output file format.

    Other Parameters
    ----------------
    nodal_distance : str ('line' or 'orthogonal'), optional, default 'line'
        Only if ``file_format = "tough"``. Method to calculate connection nodal distances:
        - 'line': distance between node and common face along connecting line (distance is not normal),
        - 'orthogonal' : distance between node and its orthogonal projection onto common face (shortest distance).
    material_name : dict or None, default None
        Only if ``file_format = "tough"``. Rename cell material.
    material_end : str, array_like or None, default None
        Only if ``file_format = "tough"``. Move cells to bottom of block 'ELEME' if their materials is in `material_end`.
    incon : bool, optional, default False
        Only if ``file_format = "tough"``. If `True`, initial conditions will be written in file `INCON`.
    protocol : integer, optional, default `pickle.HIGHEST_PROTOCOL`
        Only if ``file_format = "pickle"``. :mod:`pickle` protocol version.

    """
    # Check file format
    if not isinstance(filename, str):
        raise TypeError()
    fmt = (
        file_format
        if file_format
        else filetype_from_filename(filename, _extension_to_filetype)
    )

    # Call custom writer
    if fmt in _writer_map.keys():
        if fmt not in {"tough", "pickle"}:
            mesh = deepcopy(mesh)
            mesh.cell_data = {k: mesh.split(v) for k, v in mesh.cell_data.items()}
        _writer_map[fmt](filename, mesh, **kwargs)
    else:
        mesh = mesh.to_meshio()
        meshio.write(filename, mesh, file_format=file_format, **kwargs)


def read_time_series(filename):
    """
    Read time series from XDMF file.

    Parameters
    ----------
    filename : str
        Input file name.

    Returns
    -------
    list of namedtuple (type, data)
        Grid cell data.
    list of dict
        Data associated to grid points for each time step.
    list of dict
        Data associated to grid cells for each time step.
    array_like
        Time step values.

    """
    from ._common import get_meshio_version, get_new_meshio_cells

    if not isinstance(filename, str):
        raise ValueError()

    point_data, cell_data, time_steps = [], [], []
    if get_meshio_version() < (3,):
        reader = meshio.XdmfTimeSeriesReader(filename)
        points, cells = reader.read_points_cells()

        for k in range(reader.num_steps):
            t, pdata, cdata = reader.read_data(k)

            _, cdata = get_new_meshio_cells(cells, cdata)
            point_data.append(pdata)
            cell_data.append(cdata)
            time_steps.append(t)

        cells = get_new_meshio_cells(cells)
    else:
        with meshio.xdmf.TimeSeriesReader(filename) as reader:
            points, cells = reader.read_points_cells()

            for k in range(reader.num_steps):
                t, pdata, cdata = reader.read_data(k)
                point_data.append(pdata)
                cell_data.append(cdata)
                time_steps.append(t)

    # Concatenate cell data arrays
    for cdata in cell_data:
        for k in cdata.keys():
            cdata[k] = numpy.concatenate(cdata[k])

    return points, cells, point_data, cell_data, time_steps


def write_time_series(
    filename,
    points,
    cells,
    point_data=None,
    cell_data=None,
    time_steps=None,
):
    """
    Write time series given points and cells data.

    Parameters
    ----------
    filename : str
        Output file name.
    points : ndarray
        Grid points array.
    cells : list of namedtuple (type, data)
        Grid cell data.
    point_data : list of dict or None, optional, default None
        Data associated to grid points for each time step.
    cell_data : list of dict or None, optional, default None
        Data associated to grid cells for each time step.
    time_steps : array_like, optional, default None
        Time step values.

    """
    from ._common import get_meshio_version, get_old_meshio_cells

    if not isinstance(filename, str):
        raise TypeError()
    if point_data is not None and not isinstance(point_data, (list, tuple)):
        raise TypeError()
    if cell_data is not None and not isinstance(cell_data, (list, tuple)):
        raise TypeError()
    if time_steps is not None and not isinstance(
        time_steps, (list, tuple, numpy.ndarray)
    ):
        raise TypeError()

    if not (point_data or cell_data):
        raise ValueError("Provide at least point_data or cell_data.")
    else:
        nt = len(point_data) if point_data else len(cell_data)

    if point_data and len(point_data) != nt:
        raise ValueError("Inconsistent number of point data.")
    if cell_data and len(cell_data) != nt:
        raise ValueError("Inconsistent number of cell data.")
    if time_steps is not None and len(time_steps) != nt:
        raise ValueError("Inconsistent number of time steps.")

    point_data = point_data if point_data else [{}] * nt
    cell_data = cell_data if cell_data else [{}] * nt
    time_steps = time_steps if time_steps is not None else list(range(nt))

    # Split cell data arrays
    sizes = numpy.cumsum([len(c.data) for c in cells[:-1]])
    cell_data = [
        {k: numpy.split(v, sizes) for k, v in cdata.items()} for cdata in cell_data
    ]

    # Sort data with time steps
    idx = numpy.argsort(time_steps)
    point_data = [point_data[i] for i in idx]
    cell_data = [cell_data[i] for i in idx]
    time_steps = [time_steps[i] for i in idx]

    # Write XDMF
    def write_data(writer, points, cells, point_data, cell_data, time_steps):
        writer.write_points_cells(points, cells)
        for tstep, pdata, cdata in zip(time_steps, point_data, cell_data):
            writer.write_data(tstep, point_data=pdata, cell_data=cdata)

    if get_meshio_version() < (3,):
        writer = meshio.XdmfTimeSeriesWriter(filename)
        tmp = [get_old_meshio_cells(cells, cdata) for cdata in cell_data]
        cells = tmp[0][0]
        cell_data = [cell[1] for cell in tmp]
        write_data(writer, points, cells, point_data, cell_data, time_steps)
    else:
        with meshio.xdmf.TimeSeriesWriter(filename) as writer:
            write_data(writer, points, cells, point_data, cell_data, time_steps)



import numpy

from ._structured_grid import structured_grid

__all__ = [
    "voxelize",
]


def voxelize(points, origin, layer=False, material="dfalt"):
    """
    Generate a 3D non-uniform structured grid from cloud points.

    Input points must form a non-uniform structured grid.

    Parameters
    ----------
    points : array_like
        Cooordinates of points.
    origin : array_like
        Origin point coordinate.
    layer : bool, optional, default False
        If `True`, mesh will be generated by layers.
    material : str, optional, default 'dfalt'
        Default material name.

    Returns
    -------
    toughio.Mesh
        Output non-uniform structured mesh.

    """

    def voronoi1d(x, x0, xstr):
        """1D Voronoi tessellation."""
        if x.size > 1:
            x = numpy.sort(x)
            if x0 >= x[0]:
                raise ValueError(
                    "{}-coordinate of origin must be lower than {:.3f}, got {:.3f}".format(
                        xstr, x[0], x0
                    )
                )

            vor = [x0]
            for xx in x:
                vor.append(2 * xx - vor[-1])
        else:
            vor = [x0, x0 + 1.0]

        return vor

    if not isinstance(points, (list, tuple, numpy.ndarray)):
        raise TypeError()
    if numpy.ndim(points) != 2:
        raise ValueError()

    x, y, z = numpy.transpose(points)
    x = numpy.unique(x)
    y = numpy.unique(y)
    z = numpy.unique(z)
    if x.size * y.size * z.size != len(points):
        raise ValueError("Input points do not form a structured grid.")

    x0, y0, z0 = origin
    x = voronoi1d(x, x0, "x")
    y = voronoi1d(y, y0, "y")
    z = voronoi1d(z, z0, "z")
    dx = numpy.diff(x)
    dy = numpy.diff(y)
    dz = numpy.diff(z)
    mesh = structured_grid(dx, dy, dz, origin=origin, layer=layer, material=material)

    return mesh



from ._co2tab import co2tab
from ._export import export
from ._extract import extract
from ._merge import merge
from ._save2incon import save2incon

__all__ = [
    "co2tab",
    "export",
    "extract",
    "merge",
    "save2incon",
]



import os
import sys
from copy import deepcopy

import helpers
import numpy
import pytest

import toughio

output_ref = {
    "element": {
        0: {
            "PRES": 9641264.130638,
            "TEMP": 149.62999493,
        },
        -1: {
            "PRES": 635804.12294844,
            "TEMP": 142.89449866,
        },
    },
    "connection": {
        0: {
            "HEAT": -1.64908253e-07,
            "FLOW": -2.85760551e-13,
        },
        -1: {
            "HEAT": -5.54750914e-08,
            "FLOW": -4.68234504e-14,
        },
    },
}


def test_print():
    print(helpers.hex_mesh)


def test_extrude_to_3d():
    # Create 2D mesh
    dx = numpy.random.rand(10)
    dy = numpy.random.rand(5)
    mesh_ref = toughio.meshmaker.structured_grid(dx, dy)
    mesh_ref.point_data["a"] = numpy.random.rand(mesh_ref.n_points)
    mesh_ref.point_data["b"] = numpy.random.rand(mesh_ref.n_points, 3)
    mesh_ref.cell_data["c"] = numpy.random.rand(mesh_ref.n_cells)
    mesh_ref.cell_data["d"] = numpy.random.rand(mesh_ref.n_cells, 3)

    # Extrude mesh to 3D
    mesh = mesh_ref.extrude_to_3d(numpy.random.rand(5), axis=2, inplace=False)

    assert mesh.n_points == 6 * mesh_ref.n_points
    assert mesh.n_cells == 5 * mesh_ref.n_cells
    assert mesh.point_data["a"].shape == (6 * mesh_ref.n_points,)
    assert mesh.point_data["b"].shape == (6 * mesh_ref.n_points, 3)
    assert mesh.cell_data["c"].shape == (5 * mesh_ref.n_cells,)
    assert mesh.cell_data["d"].shape == (5 * mesh_ref.n_cells, 3)

    for v in mesh.point_data["a"].reshape((6, mesh.n_points // 6)):
        assert numpy.allclose(mesh_ref.point_data["a"], v)

    for v in mesh.point_data["b"].reshape((6, mesh.n_points // 6, 3)):
        assert numpy.allclose(mesh_ref.point_data["b"], v)

    for v in mesh.cell_data["c"].reshape((5, mesh.n_cells // 5)):
        assert numpy.allclose(mesh_ref.cell_data["c"], v)

    for v in mesh.cell_data["d"].reshape((5, mesh.n_cells // 5, 3)):
        assert numpy.allclose(mesh_ref.cell_data["d"], v)


def test_prune_duplicates():
    # Create mesh with duplicate points and cells
    points = numpy.array(
        [
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [1.0, 1.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
            [1.0, 0.0, 1.0],
            [1.0, 1.0, 1.0],
            [0.0, 1.0, 1.0],
            [2.0, 0.0, 0.0],
            [2.0, 1.0, 0.0],
            [2.0, 0.0, 1.0],
            [2.0, 1.0, 1.0],
            [2.5, 0.5, 0.5],
            [2.5, 0.5, 0.5],
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [1.0, 1.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
            [1.0, 0.0, 1.0],
            [1.0, 1.0, 1.0],
            [0.0, 1.0, 1.0],
        ]
    )
    cells = [
        ("tetra", numpy.array([[8, 12, 9, 10], [10, 13, 9, 11]])),
        (
            "wedge",
            numpy.array(
                [[1, 10, 5, 2, 11, 6], [1, 10, 5, 2, 11, 6], [1, 8, 10, 2, 9, 11]]
            ),
        ),
        (
            "hexahedron",
            numpy.array([[0, 1, 2, 3, 4, 5, 6, 7], [14, 15, 16, 17, 18, 19, 20, 21]]),
        ),
    ]
    mesh = toughio.Mesh(points, cells)
    mesh.point_data["points"] = numpy.random.rand(mesh.n_points)
    mesh.cell_data["cells"] = numpy.random.rand(mesh.n_cells)

    # Remove duplicate points and cells
    mesh.prune_duplicates()

    assert mesh.n_points == 13
    assert mesh.n_cells == 5
    assert mesh.point_data["points"].size == 13
    assert mesh.cell_data["cells"].size == 5


@pytest.mark.parametrize(
    "mesh_ref, from_, to_",
    [
        (helpers.tet_mesh, toughio.from_meshio, "to_meshio"),
        (helpers.tet_mesh, toughio.from_pyvista, "to_pyvista"),
        (helpers.hex_mesh, toughio.from_meshio, "to_meshio"),
        (helpers.hex_mesh, toughio.from_pyvista, "to_pyvista"),
    ],
)
def test_from_to(mesh_ref, from_, to_):
    mesh = from_(getattr(mesh_ref, to_)())

    helpers.allclose_mesh(mesh_ref, mesh)


@pytest.mark.parametrize(
    "filename, file_type, time_step",
    [
        ("OUTPUT_ELEME.csv", "element", 0),
        ("OUTPUT_ELEME.csv", "element", -1),
        ("OUTPUT_CONNE.csv", "connection", 0),
        ("OUTPUT_CONNE.csv", "connection", -1),
    ],
)
def test_read_output(filename, file_type, time_step):
    this_dir = os.path.dirname(os.path.abspath(__file__))
    mesh_filename = os.path.join(this_dir, "support_files", "outputs", "mesh.pickle")
    mesh = toughio.read_mesh(mesh_filename)
    filename = os.path.join(this_dir, "support_files", "outputs", filename)
    mesh.read_output(filename, time_step=time_step)

    for k, v in output_ref[file_type][time_step].items():
        assert numpy.allclose(v, mesh.cell_data[k].mean())


def test_add_point_data():
    mesh = deepcopy(helpers.hybrid_mesh)
    data = numpy.random.rand(mesh.n_points)
    mesh.add_point_data("a", data)

    assert numpy.allclose(data, mesh.point_data["a"])


def test_add_cell_data():
    mesh = deepcopy(helpers.hybrid_mesh)
    data = numpy.random.rand(mesh.n_cells)
    mesh.add_cell_data("a", data)

    assert numpy.allclose(data, mesh.cell_data["a"])


@pytest.mark.parametrize("bool_cells", [False, True])
def test_set_material(bool_cells):
    dx = numpy.ones(10)
    dy = numpy.ones(10)
    dz = numpy.ones(10)
    mesh = toughio.meshmaker.structured_grid(dx, dy, dz, origin=numpy.zeros(3))

    idx = mesh.filter.box(4.0, 4.0, 4.0, 2.0, 2.0, 2.0)
    if bool_cells:
        cells = numpy.zeros(mesh.n_cells, dtype=bool)
        cells[idx] = True
    else:
        cells = idx
    mesh.set_material("test", cells)

    assert (mesh.materials == "test").sum() == 8


def test_cell_data_to_point_data():
    mesh = deepcopy(helpers.hybrid_mesh)
    data = numpy.ones(mesh.n_cells)
    mesh.add_cell_data("a", data)
    mesh.cell_data_to_point_data()

    assert numpy.allclose(numpy.ones(mesh.n_points), mesh.point_data["a"])
    assert "a" not in mesh.cell_data.keys()


def test_point_data_to_cell_data():
    mesh = deepcopy(helpers.hybrid_mesh)
    data = numpy.ones(mesh.n_points)
    mesh.add_point_data("a", data)
    mesh.point_data_to_cell_data()

    assert numpy.allclose(numpy.ones(mesh.n_cells), mesh.cell_data["a"])
    assert "a" not in mesh.point_data.keys()


def test_near():
    dx = numpy.ones(3)
    dy = numpy.ones(3)
    dz = numpy.ones(3)
    mesh = toughio.meshmaker.structured_grid(dx, dy, dz, origin=numpy.zeros(3))

    assert mesh.near((1.5, 1.5, 1.5)) == 13


@pytest.mark.skipif(sys.version_info < (3, 6), reason="Order of keys in dictionary")
@pytest.mark.parametrize("num_pvars", [4, 6])
def test_write_tough(num_pvars):
    this_dir = os.path.dirname(os.path.abspath(__file__))
    filename = os.path.join(this_dir, "support_files", "all_cell_types.f3grid")
    mesh_ref = toughio.read_mesh(filename)

    tempdir = helpers.tempdir()
    mesh_filename = os.path.join(tempdir, "MESH")
    incon_filename = os.path.join(tempdir, "INCON")

    pvars = numpy.random.rand(mesh_ref.n_cells, num_pvars)
    mesh_ref.add_cell_data("initial_condition", pvars)

    bcond = (mesh_ref.centers[:, 2] < 0.5).astype(int)
    mesh_ref.add_cell_data("boundary_condition", bcond)

    mesh_ref.write_tough(mesh_filename, incon=True)
    mesh = toughio.read_input(mesh_filename, file_format="tough")
    incon = toughio.read_input(incon_filename, file_format="tough")

    volumes = [v["volume"] for v in mesh["elements"].values()]
    volumes = [v if v < 1.0e20 else v * 1.0e-50 for v in volumes]
    assert numpy.allclose(mesh_ref.volumes, volumes, atol=1.0e-6)

    centers = [v["center"] for v in mesh["elements"].values()]
    assert numpy.allclose(mesh_ref.centers, centers, atol=1.0e-3)

    pvars = numpy.row_stack([v["values"] for v in incon["initial_conditions"].values()])
    assert numpy.allclose(mesh_ref.cell_data["initial_condition"], pvars)



from __future__ import division

import logging
from copy import deepcopy
from functools import wraps

import numpy

dtypes = {
    "PARAMETERS": {
        "title": "str_array_like",
        "eos": "str",
        "n_component": "int",
        "n_phase": "int",
        "n_component_mass": "int",
        "flac": "dict",
        "isothermal": "bool",
        "start": "bool",
        "nover": "bool",
        "rocks": "dict",
        "rocks_order": "array_like",
        "options": "dict",
        "extra_options": "dict",
        "more_options": "dict",
        "selections": "dict",
        "solver": "dict",
        "generators": "dict",
        "times": "scalar_array_like",
        "element_history": "array_like",
        "connection_history": "array_like",
        "generator_history": "array_like",
        "diffusion": "array_like",
        "output": "dict",
        "elements": "dict",
        "elements_order": "array_like",
        "coordinates": "bool",
        "connections": "dict",
        "connections_order": "array_like",
        "initial_conditions": "dict",
        "initial_conditions_order": "array_like",
        "default": "dict",
    },
    "ROCKS": {
        "density": "scalar",
        "porosity": "scalar",
        "permeability": "scalar_array_like",
        "conductivity": "scalar",
        "specific_heat": "scalar",
        "compressibility": "scalar",
        "expansivity": "scalar",
        "conductivity_dry": "scalar",
        "tortuosity": "scalar",
        "klinkenberg_parameter": "scalar",
        "distribution_coefficient_3": "scalar",
        "distribution_coefficient_4": "scalar",
        "initial_condition": "array_like",
        "relative_permeability": "dict",
        "capillarity": "dict",
        "permeability_model": "dict",
        "equivalent_pore_pressure": "dict",
    },
    "FLAC": {"creep": "bool", "porosity_model": "int", "version": "int"},
    "MODEL": {"id": "int", "parameters": "array_like"},
    "PARAM": {
        "n_iteration": "int",
        "n_cycle": "int",
        "n_second": "int",
        "n_cycle_print": "int",
        "verbosity": "int",
        "temperature_dependence_gas": "scalar",
        "effective_strength_vapor": "scalar",
        "t_ini": "scalar",
        "t_max": "scalar",
        "t_steps": "scalar_array_like",
        "t_step_max": "scalar",
        "t_reduce_factor": "scalar",
        "gravity": "scalar",
        "mesh_scale_factor": "scalar",
        "eps1": "scalar",
        "eps2": "scalar",
        "w_upstream": "scalar",
        "w_newton": "scalar",
        "derivative_factor": "scalar",
    },
    "MOP": {i + 1: "int" for i in range(24)},
    "MOMOP": {i + 1: "int" for i in range(40)},
    "SELEC": {"integers": "dict", "floats": "array_like"},
    "SOLVR": {
        "method": "int",
        "z_precond": "str",
        "o_precond": "str",
        "rel_iter_max": "scalar",
        "eps": "scalar",
    },
    "GENER": {
        "name": "str_array_like",
        "nseq": "scalar_array_like",
        "nadd": "scalar_array_like",
        "nads": "scalar_array_like",
        "type": "str_array_like",
        "times": "scalar_array_like",
        "rates": "scalar_array_like",
        "specific_enthalpy": "scalar_array_like",
        "layer_thickness": "scalar_array_like",
    },
    "OUTPU": {"format": "str", "variables": "dict"},
    "ELEME": {
        "nseq": "int",
        "nadd": "int",
        "material": "str_int",
        "volume": "scalar",
        "heat_exchange_area": "scalar",
        "permeability_modifier": "scalar",
        "center": "array_like",
    },
    "CONNE": {
        "nseq": "int",
        "nadd": "array_like",
        "permeability_direction": "int",
        "nodal_distances": "array_like",
        "interface_area": "scalar",
        "gravity_cosine_angle": "scalar",
        "radiant_emittance_factor": "scalar",
    },
    "INCON": {"porosity": "scalar", "userx": "array_like", "values": "array_like"},
}


str_to_dtype = {
    "int": (int, numpy.int32, numpy.int64),
    "float": (float, numpy.float32, numpy.float64),
    "str": (str,),
    "bool": (bool,),
    "str_int": (str, int, numpy.int32, numpy.int64),
    "array_like": (list, tuple, numpy.ndarray),
    "dict": (dict,),
    "scalar": (int, float, numpy.int32, numpy.int64, numpy.float32, numpy.float64),
    "scalar_array_like": (
        int,
        float,
        list,
        tuple,
        numpy.int32,
        numpy.int64,
        numpy.float32,
        numpy.float64,
        numpy.ndarray,
    ),
    "str_array_like": (str, list, tuple, numpy.ndarray),
}


def block(keyword, multi=False, noend=False):
    """Decorate block writing functions."""

    def decorator(func):
        from ._common import header

        @wraps(func)
        def wrapper(*args, **kwargs):
            head_fmt = "{:5}{}" if noend else "{:5}{}\n"
            out = [head_fmt.format(keyword, header)]
            out += func(*args, **kwargs)
            out += ["\n"] if multi else []

            return out

        return wrapper

    return decorator


def check_parameters(input_types, keys=None, is_list=False):
    """Decorate function to check input parameters."""

    def _check_parameters(params, keys=None):
        for k, v in params.items():
            # Check whether parameters contain unknown keys
            # Log error if it does and skip
            if k not in input_types.keys():
                logging.warning(
                    "Unknown key '{}'{}. Skipping.".format(
                        k, " in {}".format(keys) if keys else ""
                    )
                )
                continue

            # Check input types
            input_type = str_to_dtype[input_types[k]]
            if not (v is None or isinstance(v, input_type)):
                raise TypeError(
                    "Invalid type for parameter '{}' {}(expected {}).".format(
                        k,
                        "in {} ".format(keys) if keys else "",
                        input_types[k],
                    )
                )

    keys = [keys] if isinstance(keys, str) else keys

    def decorator(func, *args, **kwargs):
        @wraps(func)
        def wrapper(parameters, *args, **kwargs):
            if not keys:
                _check_parameters(parameters)
            else:
                params = deepcopy(parameters[keys[0]])
                keys_str = "['{}']".format(keys[0])
                if is_list:
                    for k, v in params.items():
                        tmp = keys_str
                        tmp += "['{}']".format(k)
                        try:
                            for key in keys[1:]:
                                v = v[key]
                                tmp += "['{}']".format(key)
                            _check_parameters(v, tmp)
                        except KeyError:
                            continue
                else:
                    for key in keys[1:]:
                        params = params[key]
                        keys_str += "['{}']".format(key)
                    _check_parameters(params, keys_str)

            out = func(parameters, *args, **kwargs)

            return out

        return wrapper

    return decorator


def read_record(data, fmt):
    """Parse string to data given format."""
    token_to_type = {
        "s": str,
        "S": str,
        "d": int,
        "f": str2float,
        "e": str2float,
    }

    i = 0
    out = []
    for token in fmt.split(","):
        n = int(token[:-1].split(".")[0])
        tmp = data[i : i + n]
        tmp = tmp if token[-1] == "S" else tmp.strip()
        out.append(token_to_type[token[-1]](tmp) if tmp else None)
        i += n

    return out


def write_record(data, fmt, multi=False):
    """Return a list of record strings given format."""

    def to_str(x, fmt):
        x = "" if x is None else x
        return (
            fmt.format(x)
            if not isinstance(x, str)
            else fmt.replace("g", "").replace("e", "").format(x)
        )

    if not multi:
        data = [to_str(d, f) for d, f in zip(data, fmt)]
        out = ["{:80}\n".format("".join(data))]
    else:
        n = len(data)
        ncol = len(fmt)
        data = [
            data[ncol * i : min(ncol * i + ncol, n)]
            for i in range(int(numpy.ceil(n / ncol)))
        ]

        out = []
        for d in data:
            d = [to_str(dd, f) for dd, f in zip(d, fmt)]
            out += ["{:80}\n".format("".join(d))]

    return out


def str2float(s):
    """Convert variable string to float."""
    try:
        return float(s)
    except ValueError:
        # It's probably something like "0.0001-001"
        significand, exponent = s[:-4], s[-4:]
        return float("{}e{}".format(significand, exponent))


def prune_nones_dict(data):
    """Remove None key/value pairs from dict."""
    return {k: v for k, v in data.items() if v is not None}


def prune_nones_list(data):
    """Remove trailing None values from list."""
    return [x for i, x in enumerate(data) if any(xx is not None for xx in data[i:])]



from ._base import BaseCapillarity

__all__ = [
    "vanGenuchten",
]


class vanGenuchten(BaseCapillarity):

    _id = 7
    _name = "van Genuchten"

    def __init__(self, m, slr, alpha, pmax, sls):
        """
        Van Genuchten function.

        After van Genuchten (1980).

        Parameters
        ----------
        m : scalar
            Related to pore size distribution (CP(1)).
        slr : scalar
            Irreducible liquid saturation (CP(2)).
        alpha : scalar
            Inverse of capillary pressure strength 1/P0 (CP(3)).
        pmax : scalar
            Maximum pressure (CP(4)).
        sls : scalar
            Maximum liquid saturation (CP(5)).

        """
        if pmax < 0.0:
            raise ValueError()
        self.parameters = [m, slr, alpha, pmax, sls]

    def _eval(self, sl, *args):
        """Van Genuchten function."""
        m, slr, alpha, pmax, sls = args
        if sl == 1.0:
            pcap = 0.0
        else:
            if sl > slr:
                Seff = (sl - slr) / (sls - slr)
                pcap = -1.0 / abs(alpha) * (Seff ** (-1.0 / m) - 1.0) ** (1.0 - m)
            else:
                pcap = -abs(pmax)
            pcap = max(pcap, -abs(pmax))
            pcap *= (1.0 - sl) / 0.001 if sl > 0.999 else 1.0
        return pcap

    @property
    def parameters(self):
        """Return model parameters."""
        return [self._m, self._slr, self._alpha, self._pmax, self._sls]

    @parameters.setter
    def parameters(self, value):
        if len(value) != 5:
            raise ValueError()
        self._m = value[0]
        self._slr = value[1]
        self._alpha = value[2]
        self._pmax = value[3]
        self._sls = value[4]



import numpy

from .._helpers import register

__all__ = [
    "filter_",
]


def filter_(mesh, x0=None, y0=None, z0=None, dx=None, dy=None, dz=None):
    """Box filter."""
    xmin = x0 if x0 is not None else -numpy.inf
    ymin = y0 if y0 is not None else -numpy.inf
    zmin = z0 if z0 is not None else -numpy.inf

    xmax = xmin + dx if dx else numpy.inf
    ymax = ymin + dy if dy else numpy.inf
    zmax = zmin + dz if dz else numpy.inf

    x, y, z = mesh.centers.T
    mask_x = numpy.logical_and(x >= xmin, x <= xmax)
    mask_y = numpy.logical_and(y >= ymin, y <= ymax)
    mask_z = numpy.logical_and(z >= zmin, z <= zmax)
    mask = numpy.logical_and(numpy.logical_and(mask_x, mask_y), mask_z)

    return numpy.arange(mesh.n_cells)[mask]


register("box", filter_)



from ._base import BaseCapillarity

__all__ = [
    "Milly",
]


class Milly(BaseCapillarity):

    _id = 4
    _name = "Milly"

    def __init__(self, slr):
        """
        Milly's function.

        After Milly (1982).

        Parameters
        ----------
        slr : scalar
            Irreducible liquid saturation (CP(1)).

        """
        if slr < 0.0:
            raise ValueError()
        self.parameters = [slr]

    def _eval(self, sl, *args):
        (slr,) = args
        sl = max(sl, 1.001 * slr)
        fac = (
            1.0
            if sl - slr >= 0.371
            else 10 ** (2.26 * (0.371 / (sl - slr) - 1.0) ** 0.25 - 2.0)
        )
        return -97.783 * fac

    @property
    def parameters(self):
        """Return model parameters."""
        return [self._slr]

    @parameters.setter
    def parameters(self, value):
        if len(value) != 1:
            raise ValueError()
        self._slr = value[0]



import logging

import numpy

from ...__about__ import __version__ as version
from .._mesh import Mesh

__all__ = [
    "read",
    "write",
]


zone_key_to_type = {
    "T": str,
    "I": int,
    "J": int,
    "K": int,
    "N": int,
    "NODES": int,
    "E": int,
    "ELEMENTS": int,
    "F": str,
    "ET": str,
    "DATAPACKING": str,
    "ZONETYPE": str,
    "NV": int,
    "VARLOCATION": str,
}


tecplot_to_meshio_type = {
    "TRIANGLE": "triangle",
    "FETRIANGLE": "triangle",
    "QUADRILATERAL": "quad",
    "FEQUADRILATERAL": "quad",
    "TETRAHEDRON": "tetra",
    "FETETRAHEDRON": "tetra",
    "BRICK": "hexahedron",
    "FEBRICK": "hexahedron",
}


meshio_to_tecplot_type = {
    "triangle": "FETRIANGLE",
    "quad": "FEQUADRILATERAL",
    "tetra": "FETETRAHEDRON",
    "pyramid": "FEBRICK",
    "wedge": "FEBRICK",
    "hexahedron": "FEBRICK",
}


meshio_only = set(meshio_to_tecplot_type.keys())


meshio_to_tecplot_order = {
    "triangle": [0, 1, 2],
    "quad": [0, 1, 2, 3],
    "tetra": [0, 1, 2, 3],
    "pyramid": [0, 1, 2, 3, 4, 4, 4, 4],
    "wedge": [0, 1, 4, 3, 2, 2, 5, 5],
    "hexahedron": [0, 1, 2, 3, 4, 5, 6, 7],
}


meshio_to_tecplot_order_2 = {
    "triangle": [0, 1, 2, 2],
    "quad": [0, 1, 2, 3],
    "tetra": [0, 1, 2, 2, 3, 3, 3, 3],
    "pyramid": [0, 1, 2, 3, 4, 4, 4, 4],
    "wedge": [0, 1, 4, 3, 2, 2, 5, 5],
    "hexahedron": [0, 1, 2, 3, 4, 5, 6, 7],
}


meshio_type_to_ndim = {
    "triangle": 2,
    "quad": 2,
    "tetra": 3,
    "pyramid": 3,
    "wedge": 3,
    "hexahedron": 3,
}


def read(filename):
    """Read Tecplot ASCII file."""
    with open(filename, "r") as f:
        out = read_buffer(f)
    return out


def read_buffer(f):
    while True:
        line = f.readline().strip()

        if line.upper().startswith("VARIABLES"):
            variables = _read_variables(line)

        elif line.upper().startswith("ZONE"):
            # ZONE can be defined on several lines e.g.
            # ```
            # ZONE NODES = 62533, ELEMENTS = 57982
            # , DATAPACKING = BLOCK, ZONETYPE = FEQUADRILATERAL
            # , VARLOCATION = ([1-2] = NODAL, [3-7] = CELLCENTERED)
            # ```
            # is valid (and understood by ParaView and VisIt).
            lines = [line]
            i = f.tell()
            line = f.readline().strip().upper()
            while True:
                if line and not line[0].isdigit():
                    lines += [line]
                    i = f.tell()
                    line = f.readline().strip().upper()
                else:
                    f.seek(i)
                    break
            line = " ".join(lines)

            zone = _read_zone(line)
            (
                num_nodes,
                num_cells,
                zone_format,
                zone_type,
                is_cell_centered,
            ) = _parse_fezone(zone, variables)

            num_data = [num_cells if i else num_nodes for i in is_cell_centered]
            data, cells = _read_zone_data(
                f,
                sum(num_data) if zone_format == "FEBLOCK" else num_nodes,
                num_cells,
                zone_format,
            )

            break  # Only support one zone, no need to read the rest

        elif not line:
            break

    data = (
        numpy.split(numpy.concatenate(data), numpy.cumsum(num_data[:-1]))
        if zone_format == "FEBLOCK"
        else numpy.transpose(data)
    )
    data = {k: v for k, v in zip(variables, data)}

    point_data, cell_data = {}, {}
    for i, variable in zip(is_cell_centered, variables):
        if i:
            cell_data[variable] = [data[variable]]
        else:
            point_data[variable] = data[variable]

    x = "X" if "X" in point_data.keys() else "x"
    y = "Y" if "Y" in point_data.keys() else "y"
    z = "Z" if "Z" in point_data.keys() else "z" if "z" in point_data.keys() else ""
    points = numpy.column_stack((point_data.pop(x), point_data.pop(y)))
    if z:
        points = numpy.column_stack((points, point_data.pop(z)))
    cells = [(tecplot_to_meshio_type[zone_type], cells - 1)]

    return Mesh(points, cells, point_data, cell_data)


def _read_variables(line):
    # Gather variables in a list
    line = line.split("=")[1]
    line = [x for x in line.replace(",", " ").split()]
    variables = []

    i = 0
    while i < len(line):
        if '"' in line[i] and not (line[i].startswith('"') and line[i].endswith('"')):
            var = "{}_{}".format(line[i], line[i + 1])
            i += 1
        else:
            var = line[i]

        variables.append(var.replace('"', ""))
        i += 1

    # Check that at least X and Y are defined
    if "X" not in variables and "x" not in variables:
        raise ValueError("Variable 'X' not found")
    if "Y" not in variables and "y" not in variables:
        raise ValueError("Variable 'Y' not found")

    return variables


def _read_zone(line):
    # Gather zone entries in a dict
    line = line[5:]
    zone = {}

    # Look for zone title
    ivar = line.find('"')

    # If zone contains a title, process it and save the title
    if ivar >= 0:
        i1, i2 = ivar, ivar + line[ivar + 1 :].find('"') + 2
        zone_title = line[i1 + 1 : i2 - 1]
        line = line.replace(line[i1:i2], "PLACEHOLDER")
    else:
        zone_title = None

    # Look for VARLOCATION (problematic since it contains both ',' and '=')
    ivar = line.find("VARLOCATION")

    # If zone contains VARLOCATION, process it and remove the key/value pair
    if ivar >= 0:
        i1, i2 = line.find("("), line.find(")")
        zone["VARLOCATION"] = line[i1 : i2 + 1].replace(" ", "")
        line = line[:ivar] + line[i2 + 1 :]

    # Split remaining key/value pairs separated by '='
    line = [x for x in line.replace(",", " ").split() if x != "="]
    i = 0
    while i < len(line) - 1:
        if "=" in line[i]:
            if not (line[i].startswith("=") or line[i].endswith("=")):
                key, value = line[i].split("=")
            else:
                key = line[i].replace("=", "")
                value = line[i + 1]
                i += 1
        else:
            key = line[i]
            value = line[i + 1].replace("=", "")
            i += 1

        zone[key] = zone_key_to_type[key](value)
        i += 1

    # Add zone title to zone dict
    if zone_title:
        zone["T"] = zone_title

    return zone


def _parse_fezone(zone, variables):
    # Check that the grid is unstructured
    if "F" in zone.keys():
        if zone["F"] not in {"FEPOINT", "FEBLOCK"}:
            raise ValueError("Tecplot reader can only read finite-element type grids")
        if "ET" not in zone.keys():
            raise ValueError("Element type 'ET' not found")
        zone_format = zone.pop("F")
        zone_type = zone.pop("ET")
    elif "DATAPACKING" in zone.keys():
        if "ZONETYPE" not in zone.keys():
            raise ValueError("Zone type 'ZONETYPE' not found")
        zone_format = "FE" + zone.pop("DATAPACKING")
        zone_type = zone.pop("ZONETYPE")
    else:
        raise ValueError("Data format 'F' or 'DATAPACKING' not found")

    # Number of nodes
    if "N" in zone.keys():
        num_nodes = zone.pop("N")
    elif "NODES" in zone.keys():
        num_nodes = zone.pop("NODES")
    else:
        raise ValueError("Number of nodes not found")

    # Number of elements
    if "E" in zone.keys():
        num_cells = zone.pop("E")
    elif "ELEMENTS" in zone.keys():
        num_cells = zone.pop("ELEMENTS")
    else:
        raise ValueError("Number of elements not found")

    # Variable locations
    is_cell_centered = numpy.zeros(len(variables), dtype=int)
    if zone_format == "FEBLOCK":
        if "NV" in zone.keys():
            node_value = zone.pop("NV")
            is_cell_centered[node_value:] = 1
        elif "VARLOCATION" in zone.keys():
            varlocation = zone.pop("VARLOCATION")[1:-1].split(",")
            for location in varlocation:
                varrange, varloc = location.split("=")
                varloc = varloc.strip()
                if varloc == "CELLCENTERED":
                    varrange = varrange[1:-1].split("-")
                    if len(varrange) == 1:
                        i = int(varrange[0]) - 1
                        is_cell_centered[i] = 1
                    else:
                        imin = int(varrange[0]) - 1
                        imax = int(varrange[1]) - 1
                        for i in range(imin, imax + 1):
                            is_cell_centered[i] = 1

    return num_nodes, num_cells, zone_format, zone_type, is_cell_centered


def _read_zone_data(f, num_data, num_cells, zone_format):
    data, count = [], 0
    while count < num_data:
        line = f.readline().strip().split()
        if line:
            data += [[float(x) for x in line]]
            count += len(line) if zone_format == "FEBLOCK" else 1

    cells, count = [], 0
    while count < num_cells:
        line = f.readline().strip().split()
        if line:
            cells += [[[int(x) for x in line]]]
            count += 1

    return data, numpy.concatenate(cells)


def write(filename, mesh):
    """Write Tecplot ASCII file."""
    # Check cell types
    cell_types = []
    cell_blocks = []
    for ic, c in enumerate(mesh.cells):
        if c.type in meshio_only:
            cell_types.append(c.type)
            cell_blocks.append(ic)
        else:
            logging.warning(
                "Tecplot does not support cell type '{}'. Skipping cell block {}.".format(
                    c.type, ic
                )
            )

    # Define cells and zone type
    cell_types = numpy.unique(cell_types)
    if len(cell_types) == 0:
        raise AssertionError("No cell type supported by Tecplot in mesh")
    elif len(cell_types) == 1:
        # Nothing much to do except converting pyramids and wedges to hexahedra
        zone_type = meshio_to_tecplot_type[cell_types[0]]
        cells = numpy.concatenate(
            [
                mesh.cells[ic].data[:, meshio_to_tecplot_order[mesh.cells[ic].type]]
                for ic in cell_blocks
            ]
        )
    else:
        # Check if the mesh contains 2D and 3D cells
        num_dims = [meshio_type_to_ndim[mesh.cells[ic].type] for ic in cell_blocks]

        # Skip 2D cells if it does
        if len(numpy.unique(num_dims)) == 2:
            logging.warning("Mesh contains 2D and 3D cells. Skipping 2D cells.")
            cell_blocks = [ic for ic, ndim in zip(cell_blocks, num_dims) if ndim == 3]

        # Convert 2D cells to quads / 3D cells to hexahedra
        zone_type = "FEQUADRILATERAL" if num_dims[0] == 2 else "FEBRICK"
        cells = numpy.concatenate(
            [
                mesh.cells[ic].data[:, meshio_to_tecplot_order_2[mesh.cells[ic].type]]
                for ic in cell_blocks
            ]
        )

    # Define variables
    variables = ["X", "Y"]
    data = [mesh.points[:, 0], mesh.points[:, 1]]
    varrange = [3, 0]

    if mesh.points.shape[1] == 3:
        variables += ["Z"]
        data += [mesh.points[:, 2]]
        varrange[0] += 1

    for k, v in mesh.point_data.items():
        if k not in {"X", "Y", "Z", "x", "y", "z"}:
            if v.ndim == 1:
                variables += [k]
                data += [v]
                varrange[0] += 1
            elif v.ndim == 2:
                for i, vv in enumerate(v.T):
                    variables += ["{}_{}".format(k, i)]
                    data += [vv]
                    varrange[0] += 1
        else:
            logging.warning("Skipping point data '{}'.".format(k))

    if mesh.cell_data:
        varrange[1] = varrange[0] - 1
        for k, v in mesh.cell_data.items():
            if k not in {"X", "Y", "Z", "x", "y", "z"}:
                v = numpy.concatenate([v[ic] for ic in cell_blocks])
                if v.ndim == 1:
                    variables += [k]
                    data += [v]
                    varrange[1] += 1
                elif v.ndim == 2:
                    for i, vv in enumerate(v.T):
                        variables += ["{}_{}".format(k, i)]
                        data += [vv]
                        varrange[1] += 1
            else:
                logging.warning("Skipping cell data '{}'.".format(k))

    with open(filename, "w") as f:
        # Title
        f.write('TITLE = "Written by toughio v{}"\n'.format(version))

        # Variables
        variables_str = ", ".join('"{}"'.format(var) for var in variables)
        f.write("VARIABLES = {}\n".format(variables_str))

        # Zone record
        num_nodes = len(mesh.points)
        num_cells = sum(len(mesh.cells[ic].data) for ic in cell_blocks)
        f.write("ZONE NODES = {}, ELEMENTS = {},\n".format(num_nodes, num_cells))
        f.write("DATAPACKING = BLOCK, ZONETYPE = {}".format(zone_type))
        if varrange[0] <= varrange[1]:
            f.write(",\n")
            varlocation_str = (
                str(varrange[0])
                if varrange[0] == varrange[1]
                else "{}-{}".format(*varrange)
            )
            f.write("VARLOCATION = ([{}] = CELLCENTERED)\n".format(varlocation_str))
        else:
            f.write("\n")

        # Zone data
        for arr in data:
            _write_table(f, arr)

        # Cells
        cells = numpy.array(cells) + 1
        for cell in cells:
            f.write("{}\n".format(" ".join(str(c) for c in cell)))


def _write_table(f, data, ncol=20):
    nrow = len(data) // ncol
    lines = numpy.split(data, numpy.full(nrow, ncol).cumsum())
    for line in lines:
        if len(line):
            f.write("{}\n".format(" ".join(str(l) for l in line)))



import helpers
import numpy
import pytest

import toughio


def test_cylindric_grid():
    dr = numpy.array([1.0, 2.0, 3.0, 4.0])
    dz = numpy.array([1.0, 2.0, 3.0])
    mesh = toughio.meshmaker.cylindric_grid(dr, dz)

    perimeters = 2.0 * numpy.pi * dr.cumsum()
    base_areas = numpy.pi * dr.cumsum() ** 2
    surface_areas = perimeters * dz.sum()
    section_areas = dr.sum() * dz.sum()

    volumes = mesh.volumes.sum()
    assert numpy.allclose(volumes, base_areas[-1] * dz.sum())

    face_areas_top = numpy.array(mesh.face_areas)[:, 0].sum()
    assert numpy.allclose(face_areas_top, base_areas[-1] * len(dz))

    face_areas_bottom = numpy.array(mesh.face_areas)[:, 1].sum()
    assert numpy.allclose(face_areas_bottom, base_areas[-1] * len(dz))

    face_areas_section_1 = numpy.array(mesh.face_areas)[:, 2].sum()
    assert numpy.allclose(face_areas_section_1, section_areas)

    face_areas_outer = numpy.array(mesh.face_areas)[:, 3].sum()
    assert numpy.allclose(face_areas_outer, surface_areas.sum())

    face_areas_section_2 = numpy.array(mesh.face_areas)[:, 4].sum()
    assert numpy.allclose(face_areas_section_2, section_areas)

    face_areas_inner = numpy.array(mesh.face_areas)[:, 5].sum()
    assert numpy.allclose(face_areas_inner, surface_areas[:-1].sum())


@pytest.mark.parametrize("ndim", [2, 3])
def test_structured_grid(ndim):
    dx = numpy.array([1.0, 2.0, 3.0, 4.0])
    dy = numpy.array([1.0, 2.0, 3.0])
    dz = numpy.array([1.0, 2.0]) if ndim == 3 else None

    origin = numpy.random.rand(ndim)
    mesh = toughio.meshmaker.structured_grid(dx, dy, dz, origin=origin)

    assert numpy.allclose(origin, mesh.points.min(axis=0)[:ndim])

    if ndim == 3:
        volumes_ref = dx.sum() * dy.sum() * dz.sum()
        assert numpy.allclose(volumes_ref, mesh.volumes.sum())


@pytest.mark.parametrize("ndim", [2, 3])
def test_triangulate(ndim):
    points = numpy.random.rand(100, ndim)
    mesh = toughio.meshmaker.triangulate(points)

    assert numpy.allclose(points, mesh.points)
    assert mesh.cells[0].type == ("triangle" if ndim == 2 else "tetra")


def test_voxelize():
    dx = numpy.array([1.0, 2.0, 3.0, 4.0])
    dy = numpy.array([1.0, 2.0, 3.0])
    dz = numpy.array([1.0, 2.0])
    origin = numpy.random.rand(3)
    mesh_ref = toughio.meshmaker.structured_grid(dx, dy, dz, origin=origin)

    mesh = toughio.meshmaker.voxelize(mesh_ref.centers, origin=origin)
    helpers.allclose_mesh(mesh_ref, mesh)


def test_layer():
    dr = numpy.array([1.0, 2.0, 3.0, 4.0])
    dz = numpy.array([1.0, 2.0, 3.0])

    mesh = toughio.meshmaker.cylindric_grid(dr, dz, layer=False)
    mesh_layer = toughio.meshmaker.cylindric_grid(dr, dz, layer=True)

    point = 3.5, 0.0, 1.5
    i1 = mesh.near(point)
    i2 = mesh_layer.near(point)

    assert numpy.allclose(mesh.volumes[i1], mesh_layer.volumes[i2])
    assert numpy.allclose(mesh.face_areas[i1], mesh_layer.face_areas[i2])



import collections
from copy import deepcopy

import meshio
import numpy

from ._common import get_meshio_version, get_new_meshio_cells, get_old_meshio_cells
from ._filter import MeshFilter
from ._properties import (
    _connections,
    _face_areas,
    _face_normals,
    _faces,
    _materials,
    _qualities,
    _volumes,
)

__all__ = [
    "CellBlock",
    "Mesh",
    "from_meshio",
    "from_pyvista",
]


CellBlock = collections.namedtuple("CellBlock", ["type", "data"])


class Mesh(object):
    def __init__(
        self,
        points,
        cells,
        point_data=None,
        cell_data=None,
        field_data=None,
        point_sets=None,
        cell_sets=None,
    ):
        """
        Unstructured toughio mesh.

        This class is updated following the latest :mod:`meshio` version and
        brings backward compatibility with its previous versions.

        Parameters
        ----------
        points : array_like (n_points, 3)
            Cooordinates of points.
        cells : list of tuples (cell_type, data)
            Connectivity of cells.
        point_data : dict or None, optional, default None
            Point data arrays.
        cell_data : dict or None, optional, default None
            Cell data arrays.
        field_data : dict or None, optional, default None
            Field data names.
        point_sets : dict or None, optional, default None
            Sets of points.
        cell_sets : dict or None, optional, default None
            Sets of cells.

        """
        self.points = points
        self.cells = cells
        self.point_data = point_data if point_data else {}
        self.cell_data = cell_data if cell_data else {}
        self.field_data = field_data if field_data else {}
        self.point_sets = point_sets if point_sets else {}
        self.cell_sets = cell_sets if cell_sets else {}
        self.label_length = None

    def __repr__(self):
        """Represent a :class:`toughio.Mesh`."""
        lines = [
            "<toughio mesh object>",
            "  Number of points: {}".format(len(self.points)),
        ]
        if len(self.cells) > 0:
            lines.append("  Number of cells:")
            for tpe, elems in self.cells:
                lines.append("    {}: {}".format(tpe, len(elems)))
        else:
            lines.append("  No cells.")

        if self.point_sets:
            lines.append("  Point sets: {}".format(", ".join(self.point_sets.keys())))

        if self.point_data:
            lines.append("  Point data: {}".format(", ".join(self.point_data.keys())))

        if self.cell_data:
            lines.append("  Cell data: {}".format(", ".join(self.cell_data.keys())))

        return "\n".join(lines)

    def extrude_to_3d(self, height=1.0, axis=2, inplace=True):
        """
        Convert a 2D mesh to 3D by extruding cells along given axis.

        Parameters
        ----------
        height : scalar or array_like, optional, default 1.0
            Height of extrusion.
        axis : int (0, 1 or 2), optional, default 2
            Axis along which extrusion is performed.
        inplace : bool, optional, default True
            If `False`, return a new :class:`toughio.Mesh`.

        Returns
        -------
        toughio.Mesh
            Extruded mesh (only if ``inplace == False``).

        """
        if axis not in [0, 1, 2]:
            raise ValueError("axis must be 0, 1 or 2.")
        mesh = self if inplace else deepcopy(self)
        height = [height] if numpy.ndim(height) == 0 else height

        npts, nh = len(mesh.points), len(height)
        if mesh.points.shape[1] == 3:
            if len(set(mesh.points[:, axis])) != 1:
                raise ValueError("Cannot extrude mesh along axis {}.".format(axis))
        else:
            mesh.points = numpy.column_stack((mesh.points, numpy.zeros(npts)))
            if axis != 2:
                mesh.points[:, [axis, 2]] = mesh.points[:, [2, axis]]

        extra_points = numpy.array(mesh.points)
        for h in height:
            extra_points[:, axis] += h
            mesh.points = numpy.vstack((mesh.points, extra_points))
        for k, v in mesh.point_data.items():
            mesh.point_data[k] = (
                numpy.tile(v, nh + 1)
                if numpy.ndim(v) == 1
                else numpy.tile(v, (nh + 1, 1))
            )

        extruded_types = {
            "triangle": "wedge",
            "quad": "hexahedron",
        }
        cells = []
        cell_data = {k: mesh.split(v) for k, v in mesh.cell_data.items()}
        for ic, c in enumerate(mesh.cells):
            if c.type in extruded_types.keys():
                extruded_type = extruded_types[c.type]
                nr, nc = c.data.shape
                cell = CellBlock(extruded_type, numpy.tile(c.data, (nh, 2)))
                for i in range(nh):
                    ibeg, iend = i * nr, (i + 1) * nr
                    cell.data[ibeg:iend, :nc] += i * npts
                    cell.data[ibeg:iend, nc:] += (i + 1) * npts
                cells.append(cell)

                for k, v in cell_data.items():
                    v[ic] = (
                        numpy.tile(v[ic], nh)
                        if numpy.ndim(v[ic]) == 1
                        else numpy.tile(v[ic], (nh, 1))
                    )

        mesh.cells = cells
        mesh.cell_data = {k: numpy.concatenate(v) for k, v in cell_data.items()}

        if mesh.field_data:
            for k in mesh.field_data.keys():
                mesh.field_data[k][1] = 3

        if not inplace:
            return mesh

    def prune_duplicates(self, inplace=True):
        """
        Delete duplicate points and cells.

        Parameters
        ----------
        inplace : bool, optional, default True
            If `False`, return a new :class:`toughio.Mesh`.

        Returns
        -------
        toughio.Mesh
            Pruned mesh (only if ``inplace == False``).

        Note
        ----
        Does not preserve points order from original array in mesh.

        """
        mesh = self if inplace else deepcopy(self)
        cells = [[c.type, c.data] for c in mesh.cells]

        # Prune duplicate points
        unique_points, pind, pinv = numpy.unique(
            mesh.points,
            axis=0,
            return_index=True,
            return_inverse=True,
        )
        if len(unique_points) < len(mesh.points):
            mesh.points = unique_points
            for k, v in mesh.point_data.items():
                mesh.point_data[k] = v[pind]
            for ic, (k, v) in enumerate(cells):
                cells[ic][1] = pinv[v]

        # Prune duplicate cells
        cell_data = {k: mesh.split(v) for k, v in mesh.cell_data.items()}
        for ic, (k, v) in enumerate(cells):
            vsort = numpy.sort(v, axis=1)
            _, order = numpy.unique(vsort, axis=0, return_index=True)
            cells[ic][1] = v[order]
            for kk, vv in cell_data.items():
                cell_data[kk][ic] = vv[ic][order]
        mesh.cells = cells
        mesh.cell_data = {k: numpy.concatenate(v) for k, v in cell_data.items()}

        if not inplace:
            return mesh

    def split(self, arr):
        """
        Split input array into subarrays for each cell block in mesh.

        Parameters
        ----------
        arr : array_like
            Input array.

        Returns
        -------
        list of array_like
            List of subarrays.

        """
        if len(arr) != self.n_cells:
            raise ValueError()
        sizes = numpy.cumsum([len(c.data) for c in self.cells])

        return numpy.split(numpy.asarray(arr), sizes[:-1])

    def to_meshio(self):
        """
        Convert mesh to :class:`meshio.Mesh`.

        Returns
        -------
        meshio.Mesh
            Output mesh.

        """
        keys = ["points", "point_data", "field_data"]
        kwargs = {key: getattr(self, key) for key in keys}

        version = get_meshio_version()
        cell_data = {k: self.split(v) for k, v in self.cell_data.items()}
        if version[0] >= 4:
            kwargs.update(
                {
                    "cells": self.cells,
                    "cell_data": cell_data,
                    "point_sets": self.point_sets,
                    "cell_sets": self.cell_sets,
                }
            )
        else:
            cells, cell_data = get_old_meshio_cells(self.cells, cell_data)
            kwargs.update(
                {
                    "cells": cells,
                    "cell_data": cell_data,
                    "node_sets": self.point_sets,
                }
            )

        return meshio.Mesh(**kwargs)

    def to_pyvista(self):
        """
        Convert mesh to :class:`pyvista.UnstructuredGrid`.

        Returns
        -------
        pyvista.UnstructuredGrid
            Output mesh.

        """
        try:
            import pyvista
            import vtk

            from ._common import meshio_to_vtk_type, vtk_type_to_numnodes

            VTK9 = vtk.vtkVersion().GetVTKMajorVersion() >= 9
        except ImportError:
            raise ImportError(
                "Converting to pyvista.UnstructuredGrid requires pyvista to be installed."
            )

        # Extract cells from toughio.Mesh object
        offset = []
        cells = []
        cell_type = []
        next_offset = 0
        for c in self.cells:
            vtk_type = meshio_to_vtk_type[c.type]
            numnodes = vtk_type_to_numnodes[vtk_type]
            cells.append(
                numpy.hstack((numpy.full((len(c.data), 1), numnodes), c.data)).ravel()
            )
            cell_type += [vtk_type] * len(c.data)
            if not VTK9:
                offset += [next_offset + i * (numnodes + 1) for i in range(len(c.data))]
                next_offset = offset[-1] + numnodes + 1

        # Create pyvista.UnstructuredGrid object
        points = self.points
        if points.shape[1] == 2:
            points = numpy.hstack((points, numpy.zeros((len(points), 1))))

        if VTK9:
            mesh = pyvista.UnstructuredGrid(
                numpy.concatenate(cells),
                numpy.array(cell_type),
                numpy.array(points, numpy.float64),
            )
        else:
            mesh = pyvista.UnstructuredGrid(
                numpy.array(offset),
                numpy.concatenate(cells),
                numpy.array(cell_type),
                numpy.array(points, numpy.float64),
            )

        # Set point data
        mesh.point_arrays.update(
            {k: numpy.array(v, numpy.float64) for k, v in self.point_data.items()}
        )

        # Set cell data
        mesh.cell_arrays.update(self.cell_data)

        return mesh

    def write_tough(self, filename="MESH", **kwargs):
        """
        Write TOUGH `MESH` file.

        Parameters
        ----------
        filename : str, optional, default 'MESH'
            Output file name.

        Other Parameters
        ----------------
        nodal_distance : str ('line' or 'orthogonal'), optional, default 'line'
            Method to calculate connection nodal distances:
            - 'line': distance between node and common face along connecting
            line (distance is not normal),
            - 'orthogonal' : distance between node and its orthogonal
            projection onto common face (shortest distance).
        material_name : dict or None, default None
            Rename cell material.
        material_end : str, array_like or None, default None
            Move cells to bottom of block 'ELEME' if their materials is in `material_end`.
        incon : bool, optional, default False
            If `True`, initial conditions will be written in file `INCON`.

        """
        self.write(filename, file_format="tough", **kwargs)

    def write_incon(self, filename="INCON"):
        """
        Write TOUGH `INCON` file.

        Parameters
        ----------
        filename : str, optional, default 'INCON'
            Output file name.

        Note
        ----
        Mostly useful to restart a simulation with other initial conditions but with the same
        mesh.

        """
        from .tough._tough import check_incon, init_incon, write_incon

        primary_variables, porosities, permeabilities = init_incon(self)
        incon = check_incon(
            True, primary_variables, porosities, permeabilities, self.n_cells
        )

        if incon:
            write_incon(
                filename,
                self.labels,
                primary_variables,
                porosities,
                permeabilities,
            )

    def read_output(self, file_or_output, time_step=-1, connection=False):
        """
        Import TOUGH results to the mesh.

        Parameters
        ----------
        file_or_output : str, namedtuple or list of namedtuple
            Input file name or output data.
        time_step : int, optional, default -1
            Data for given time step to import. Default is last time step.
        connection : bool, optional, default False
            Only for standard TOUGH output file. If `True`, read data related to connections.

        """
        from .. import read_output
        from .._io.output._common import Output, reorder_labels

        if not isinstance(file_or_output, (str, list, tuple, Output)):
            raise TypeError()
        if not isinstance(time_step, int):
            raise TypeError()

        if isinstance(file_or_output, str):
            out = read_output(file_or_output, connection=connection)
        else:
            out = file_or_output

        if not isinstance(out, Output):
            if not (-len(out) <= time_step < len(out)):
                raise ValueError()
            out = out[time_step]

        if out.type == "element":
            if len(out.labels) != self.n_cells:
                raise ValueError()

            out = reorder_labels(out, self.labels)
            self.cell_data.update(out.data)
        elif out.type == "connection":
            centers = self.centers
            labels_map = {k: v for v, k in enumerate(self.labels)}

            data = {
                k: [[[0.0, 0.0, 0.0]] for _ in range(self.n_cells)]
                for k in out.data.keys()
            }
            for i, (label1, label2) in enumerate(out.labels):
                i1, i2 = labels_map[label1], labels_map[label2]
                line = centers[i1] - centers[i2]
                line /= numpy.linalg.norm(line)

                for k, v in out.data.items():
                    iv = i1 if v[i] > 0.0 else i2
                    data[k][iv].append(v[i] * line)

            data = {
                k: numpy.array([numpy.sum(vv, axis=0) for vv in v])
                for k, v in data.items()
            }
            self.cell_data.update(data)

    def write(self, filename, file_format=None, **kwargs):
        """
        Write mesh to file.

        Parameters
        ----------
        filename : str
            Output file name.
        file_format : str or None, optional, default None
            Output file format. If `None`, it will be guessed from file's
            extension. To write TOUGH MESH, `file_format` must be specified
            as 'tough' (no specific extension exists for TOUGH MESH).

        Other Parameters
        ----------------
        nodal_distance : str ('line' or 'orthogonal'), optional, default 'line'
            Only if ``file_format = "tough"``. Method to calculate connection
            nodal distances:
            - 'line': distance between node and common face along connecting
            line (distance is not normal),
            - 'orthogonal' : distance between node and its orthogonal
            projection onto common face (shortest distance).
        material_name : dict or None, default None
            Only if ``file_format = "tough"``. Rename cell material.
        material_end : str, array_like or None, default None
            Only if ``file_format = "tough"``. Move cells to bottom of block
            'ELEME' if their materials is in `material_end`.
        incon : bool, optional, default False
            Only if ``file_format = "tough"``. If `True`, initial conditions will be written in file `INCON`.
        protocol : integer, optional, default `pickle.HIGHEST_PROTOCOL`
            Only if ``file_format = "pickle"``. :mod:`pickle` protocol version.

        """
        from ._helpers import write

        write(filename, self, file_format, **kwargs)

    def plot(self, *args, **kwargs):
        """Display mesh using :meth:`pyvista.UnstructuredGrid.plot`."""
        mesh = self.to_pyvista()
        mesh.plot(*args, **kwargs)

    def add_point_data(self, label, data):
        """
        Add a new point data array.

        Parameters
        ----------
        label : str
            Point data array name.
        data : array_like
            Point data array.

        """
        if not isinstance(label, str):
            raise TypeError()
        if not isinstance(data, (list, tuple, numpy.ndarray)):
            raise TypeError()
        if len(data) != self.n_points:
            raise ValueError()
        self.point_data[label] = numpy.asarray(data)

    def add_cell_data(self, label, data):
        """
        Add a new cell data array.

        Parameters
        ----------
        label : str
            Cell data array name.
        data : array_like
            Cell data array.

        """
        if not isinstance(label, str):
            raise TypeError()
        if not isinstance(data, (list, tuple, numpy.ndarray)):
            raise TypeError()
        if len(data) != self.n_cells:
            raise ValueError()
        self.cell_data[label] = numpy.asarray(data)

    def set_material(self, material, cells):
        """
        Set material to cells.

        Parameters
        ----------
        material : str
            Material name.
        cells : array_like
            Indices of cells or array of booleans.

        """
        ints = (int, numpy.int8, numpy.int16, numpy.int32, numpy.int64)
        cond_int = all(isinstance(c, ints) for c in cells)
        cond_bool = all(isinstance(c, (bool, numpy.bool_)) for c in cells)

        if cond_int:
            if numpy.min(cells) < 0 or numpy.max(cells) >= self.n_cells:
                raise ValueError()
        elif cond_bool:
            if len(cells) != self.n_cells:
                raise ValueError()
            cells = numpy.arange(self.n_cells)[cells]
        else:
            raise TypeError()

        if len(cells):
            data = self.cell_data["material"]
            imat = (
                self.field_data[material][0]
                if material in self.field_data.keys()
                else data.max() + 1
            )
            data[cells] = imat
            self.add_cell_data("material", data)
            self.field_data[material] = numpy.array([imat, 3])

    def cell_data_to_point_data(self):
        """Interpolate cell data to point data."""
        from ._common import interpolate_data

        points = [[] for _ in range(self.n_points)]
        weights = [[] for _ in range(self.n_points)]

        i = 0
        volumes = self.volumes
        for cellblock in self.cells:
            for cell in cellblock.data:
                for c in cell:
                    points[c].append(i)
                    weights[c].append(volumes[i])
                i += 1

        point_data = interpolate_data(self._cell_data, points, weights)
        self._point_data.update(point_data)
        self._cell_data = {}

    def point_data_to_cell_data(self):
        """Interpolate point data to cell data."""
        from ._common import interpolate_data

        cells = [c for cell in self.cells for c in cell.data]

        cell_data = interpolate_data(self._point_data, cells)
        self._cell_data.update(cell_data)
        self._point_data = {}

    def near(self, points):
        """
        Return indices of cells nearest to query points.

        Parameters
        ----------
        points : array_like
            Coordinates of points to query.

        Returns
        -------
        int or array_like
            Index of cell or indices of cells.

        """

        def distance(point, points):
            """Distance between point to list of points."""
            dp = points - point
            return numpy.einsum("ij,ij->i", dp, dp)

        ndim = numpy.ndim(points)
        if ndim == 0:
            raise TypeError()
        elif ndim == 1:
            points = numpy.array([points])
        else:
            points = numpy.asarray(points)
        if points.shape[1] != self.points.shape[1]:
            raise ValueError()

        centers = self.centers
        idx = numpy.argmin([distance(point, centers) for point in points], axis=1)
        return idx[0] if ndim == 1 else idx

    @property
    def points(self):
        """Return coordinates of points."""
        return self._points

    @points.setter
    def points(self, value):
        self._points = value

    @property
    def cells(self):
        """Return connectivity of cells."""
        if self._cells:
            return self._cells
        else:
            return [CellBlock(k, v) for k, v in self._cells_dict.items()]

    @cells.setter
    def cells(self, value):
        if isinstance(value, dict):
            self._cells = []
            self._cells_dict = value
        else:
            self._cells = [CellBlock(k, v) for k, v in value]
            self._cells_dict = {}

    @property
    def cells_dict(self):
        """Return connectivity of cells (``meshio < 4.0.0``)."""
        if self._cells:
            return get_old_meshio_cells(self._cells)
        else:
            return self._cells_dict

    @property
    def point_data(self):
        """Return point data arrays."""
        return self._point_data

    @point_data.setter
    def point_data(self, value):
        self._point_data = value

    @property
    def cell_data(self):
        """Return cell data arrays."""
        return self._cell_data

    @cell_data.setter
    def cell_data(self, value):
        self._cell_data = value

    @property
    def field_data(self):
        """Return field data names."""
        return self._field_data

    @field_data.setter
    def field_data(self, value):
        self._field_data = value

    @property
    def point_sets(self):
        """Return sets of points."""
        return self._point_sets

    @point_sets.setter
    def point_sets(self, value):
        self._point_sets = value

    @property
    def cell_sets(self):
        """Return sets of cells."""
        return self._cell_sets

    @cell_sets.setter
    def cell_sets(self, value):
        self._cell_sets = value

    @property
    def n_points(self):
        """Return number of points."""
        return len(self.points)

    @property
    def n_cells(self):
        """Return number of cells."""
        return sum(len(c.data) for c in self.cells)

    @property
    def label_length(self):
        """Return length of labels."""
        return self._label_length

    @label_length.setter
    def label_length(self, value):
        self._label_length = value

    @property
    def labels(self):
        """Return labels of cell in mesh."""
        from ._common import labeler

        return labeler(self.n_cells, self.label_length)

    @property
    def centers(self):
        """Return node centers of cell in mesh."""
        return numpy.concatenate([self.points[c.data].mean(axis=1) for c in self.cells])

    @property
    def materials(self):
        """Return materials of cell in mesh."""
        return _materials(self)

    @property
    def faces(self):
        """Return connectivity of faces of cell in mesh."""
        out = _faces(self)
        arr = numpy.full((self.n_cells, 6, 4), -1)
        for i, x in enumerate(out):
            arr[i, : len(x[0]), : x[0].shape[1]] = x[0]
            if len(x) > 1:
                arr[i, len(x[0]) : len(x[0]) + len(x[1]), : x[1].shape[1]] = x[1]
        return arr

    @property
    def face_normals(self):
        """Return normal vectors of faces in mesh."""
        return _face_normals(self)

    @property
    def face_areas(self):
        """Return areas of faces in mesh."""
        return _face_areas(self)

    @property
    def volumes(self):
        """Return volumes of cell in mesh."""
        return _volumes(self)

    @property
    def connections(self):
        """
        Return mesh connections.

        Assume conformity and that points and cells are uniquely defined in mesh.

        Note
        ----
        Only for 3D meshes and first order cells.

        """
        return _connections(self)

    @property
    def qualities(self):
        """
        Return qualities of cells in mesh.

        The quality of a cell is measured as the minimum cosine angle between the
        connection line and the interface normal vectors.

        """
        return numpy.array([numpy.min(out) for out in _qualities(self)])

    @property
    def filter(self):
        """
        Filter mesh.

        Parameters
        ----------
        filter_ : str, optional, default 'box'
            Filter method.

        Returns
        -------
        array_like
            Indices of cells filtered.

        """
        return MeshFilter(self)


def from_meshio(mesh, material="dfalt"):
    """
    Convert a :class:`meshio.Mesh` to :class:`toughio.Mesh`.

    Parameters
    ----------
    mesh : meshio.Mesh
        Input mesh.
    material : str, optional, default 'dfalt'
        Default material name.

    Returns
    -------
    toughio.Mesh
        Output mesh.

    """
    from ._helpers import get_material_key

    if not isinstance(mesh, meshio.Mesh):
        raise TypeError()
    if not isinstance(material, str):
        raise TypeError()

    version = get_meshio_version()

    if mesh.cell_data:
        if version[0] >= 4:
            cells = mesh.cells
            cell_data = mesh.cell_data
        else:
            cells, cell_data = get_new_meshio_cells(mesh.cells, mesh.cell_data)

        key = get_material_key(cell_data)
        if key:
            cell_data["material"] = cell_data.pop(key)
        cell_data = {k: numpy.concatenate(v) for k, v in cell_data.items()}
    else:
        cells = mesh.cells if version[0] >= 4 else get_new_meshio_cells(mesh.cells)
        cell_data = {}

    out = Mesh(
        points=mesh.points,
        cells=cells,
        point_data=mesh.point_data,
        cell_data=cell_data,
        field_data=mesh.field_data,
        point_sets=(
            mesh.point_sets
            if hasattr(mesh, "point_sets")
            else mesh.node_sets
            if hasattr(mesh, "node_sets")
            else None
        ),
        cell_sets=mesh.cell_sets if hasattr(mesh, "cell_sets") else None,
    )

    if "material" not in out.cell_data.keys():
        imat = (
            numpy.max([v[0] for v in mesh.field_data.values() if v[1] == 3]) + 1
            if mesh.field_data
            else 1
        )
        out.cell_data["material"] = numpy.full(out.n_cells, imat, dtype=int)
        out.field_data[material] = numpy.array([imat, 3])

    return out


def from_pyvista(mesh, material="dfalt"):
    """
    Convert a :class:`pyvista.UnstructuredGrid` to :class:`toughio.Mesh`.

    Parameters
    ----------
    mesh : pyvista.UnstructuredGrid
        Input mesh.
    material : str, optional, default 'dfalt'
        Default material name.

    Returns
    -------
    toughio.Mesh
        Output mesh.

    """
    try:
        import pyvista
        import vtk

        from ._common import vtk_to_meshio_type

        VTK9 = vtk.vtkVersion().GetVTKMajorVersion() >= 9
    except ImportError:
        raise ImportError(
            "Converting pyvista.UnstructuredGrid requires pyvista to be installed."
        )

    if not isinstance(mesh, pyvista.UnstructuredGrid):
        raise TypeError()
    if not isinstance(material, str):
        raise TypeError()

    # Copy useful arrays to avoid repeated calls to properties
    vtk_offset = mesh.offset
    vtk_cells = mesh.cells
    vtk_cell_type = mesh.celltypes

    # Check that meshio supports all cell types in input mesh
    pixel_voxel = {8, 11}  # Handle pixels and voxels
    for cell_type in numpy.unique(vtk_cell_type):
        if not (cell_type in vtk_to_meshio_type.keys() or cell_type in pixel_voxel):
            raise ValueError("toughio does not support VTK type {}.".format(cell_type))

    # Get cells
    cells = []
    c = 0
    for offset, cell_type in zip(vtk_offset, vtk_cell_type):
        numnodes = vtk_cells[offset]
        if VTK9:
            cell = vtk_cells[offset + 1 + c : offset + 1 + c + numnodes]
            c += 1
        else:
            cell = vtk_cells[offset + 1 : offset + 1 + numnodes]
        cell = (
            cell
            if cell_type not in pixel_voxel
            else cell[[0, 1, 3, 2]]
            if cell_type == 8
            else cell[[0, 1, 3, 2, 4, 5, 7, 6]]
        )
        cell_type = cell_type if cell_type not in pixel_voxel else cell_type + 1
        cell_type = (
            vtk_to_meshio_type[cell_type]
            if cell_type != 7
            else "polygon{}".format(numnodes)
        )

        if len(cells) > 0 and cells[-1][0] == cell_type:
            cells[-1][1].append(cell)
        else:
            cells.append((cell_type, [cell]))

    for k, c in enumerate(cells):
        cells[k] = (c[0], numpy.array(c[1]))

    # Get point data
    point_data = {k.replace(" ", "_"): v for k, v in mesh.point_arrays.items()}

    # Get cell data
    cell_data = {k.replace(" ", "_"): v for k, v in mesh.cell_arrays.items()}

    # Create toughio.Mesh
    out = Mesh(
        points=numpy.array(mesh.points),
        cells=cells,
        point_data=point_data,
        cell_data=cell_data,
    )

    if "material" not in out.cell_data.keys():
        imat = 1
        out.cell_data["material"] = numpy.full(out.n_cells, imat, dtype=int)
        out.field_data[material] = numpy.array([imat, 3])

    return out



from .input import read as read_input
from .input import register as register_input
from .input import write as write_input
from .output import Output
from .output import read as read_output
from .output import read_history
from .output import register as register_output
from .output import write as write_output

__all__ = [
    "Output",
    "register_input",
    "register_output",
    "read_input",
    "write_input",
    "read_output",
    "write_output",
    "read_history",
]



from __future__ import with_statement

import numpy

from ...input import tough
from .._common import Output, reorder_labels

__all__ = [
    "read",
]


def read(filename, file_type, file_format, labels_order):
    """Read SAVE."""
    parameters = tough.read(filename)

    labels = list(parameters["initial_conditions"].keys())
    variables = [v["values"] for v in parameters["initial_conditions"].values()]

    data = {"X{}".format(i + 1): x for i, x in enumerate(numpy.transpose(variables))}

    data["porosity"] = numpy.array(
        [v["porosity"] for v in parameters["initial_conditions"].values()]
    )

    userx = [
        v["userx"]
        for v in parameters["initial_conditions"].values()
        if "userx" in v.keys()
    ]
    if userx:
        data["userx"] = numpy.array(userx)

    labels_order = (
        labels_order if labels_order else parameters["initial_conditions_order"]
    )
    output = Output(file_type, file_format, None, numpy.array(labels), data)
    return reorder_labels(output, labels_order)
